'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.StructError = exports.superstruct = exports.struct = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _cloneDeep = require('lodash/cloneDeep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _componentType = require('component-type');

var _componentType2 = _interopRequireDefault(_componentType);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Default types.
 *
 * @type {Object}
 */

var DEFAULT_TYPES = {
  any: function any(v) {
    return v !== undefined;
  },
  array: function array(v) {
    return (0, _componentType2.default)(v) === 'array';
  },
  boolean: function boolean(v) {
    return (0, _componentType2.default)(v) === 'boolean';
  },
  buffer: function buffer(v) {
    return (0, _componentType2.default)(v) === 'buffer';
  },
  date: function date(v) {
    return (0, _componentType2.default)(v) === 'date';
  },
  error: function error(v) {
    return (0, _componentType2.default)(v) === 'error';
  },
  function: function _function(v) {
    return (0, _componentType2.default)(v) === 'string';
  },
  null: function _null(v) {
    return (0, _componentType2.default)(v) === 'null';
  },
  number: function number(v) {
    return (0, _componentType2.default)(v) === 'number';
  },
  object: function object(v) {
    return (0, _componentType2.default)(v) === 'object';
  },
  regexp: function regexp(v) {
    return (0, _componentType2.default)(v) === 'regexp';
  },
  string: function string(v) {
    return (0, _componentType2.default)(v) === 'string';
  },
  undefined: function undefined(v) {
    return (0, _componentType2.default)(v) === 'undefined';
  }

  /**
   * Define a struct error.
   *
   * @type {StructError}
   */

};
var StructError = function (_Error) {
  _inherits(StructError, _Error);

  function StructError(message, data) {
    _classCallCheck(this, StructError);

    data.code = message;
    data.path = data.path || [];
    var index = data.index,
        key = data.key,
        value = data.value,
        type = data.type;


    switch (message) {
      case 'element_invalid':
        message = 'Expected the element at index `' + index + '` to be of type "' + type + '", but it was `' + value + '`.';
        break;
      case 'property_invalid':
      case 'property_required':
        message = 'Expected the `' + key + '` property to be of type "' + type + '", but it was `' + value + '`.';
        break;
      case 'property_unknown':
        message = 'Unexpected `' + key + '` property that was not defined in the struct.';
        break;
      case 'value_invalid':
      case 'value_required':
        message = 'Expected a value of type "' + type + '" but received `' + value + '`.';
        break;
      default:
        throw new Error('Unknown struct error code: "' + message + '"');
    }

    var _this = _possibleConstructorReturn(this, (StructError.__proto__ || Object.getPrototypeOf(StructError)).call(this, message));

    _this.name = 'StructError';

    for (var k in data) {
      _this[k] = data[k];
    }

    Error.captureStackTrace(_this, _this.constructor);
    return _this;
  }

  return StructError;
}(Error);

/**
 * Create a struct factory from a set of `options`.
 *
 * @param {Object} options
 * @return {Function}
 */

function superstruct() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var TYPES = _extends({}, DEFAULT_TYPES, options.types || {});

  /**
   * Define a scalar struct with a `schema` type string.
   *
   * @param {String} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function scalarStruct(schema, defaults) {
    var isOptional = schema.endsWith('?');
    var type = isOptional ? schema.slice(0, -1) : schema;
    var types = type.split(/\s*\|\s*/g);

    var fns = types.map(function (t) {
      var fn = TYPES[t];

      if (typeof fn !== 'function') {
        throw new Error('No struct validator function found for type "' + t + '".');
      }

      return fn;
    });

    return function (value) {
      if (!isOptional && value === undefined) {
        throw new StructError('value_required', { type: type });
      }

      if (value !== undefined && !fns.some(function (fn) {
        return fn(value);
      })) {
        throw new StructError('value_invalid', { type: type, value: value });
      }

      return value;
    };
  }

  /**
   * Define a list struct with a `schema` array.
   *
   * @param {Array} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function listStruct(schema, defaults) {
    if (schema.length !== 1) {
      throw new Error('List structs must be defined as an array with a single element, but you passed ' + schema.length + ' elements.');
    }

    schema = schema[0];
    var fn = struct(schema);
    var type = 'array';

    return function (value) {
      if (value === undefined) {
        throw new StructError('value_required', { type: type });
      } else if ((0, _componentType2.default)(value) !== 'array') {
        throw new StructError('value_invalid', { type: type, value: value });
      }

      var ret = value.map(function (v, index) {
        try {
          return fn(v);
        } catch (e) {
          var path = [index].concat(e.path);

          switch (e.code) {
            case 'value_invalid':
              throw new StructError('element_invalid', _extends({}, e, { index: index, path: path }));
            default:
              if ('path' in e) e.path = path;
              throw e;
          }
        }
      });

      return ret;
    };
  }

  /**
   * Define an object struct with a `schema` dictionary.
   *
   * @param {Object} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function objectStruct(schema, defaults) {
    var structs = {};
    var type = 'object';

    for (var _key in schema) {
      var fn = struct(schema[_key]);
      structs[_key] = fn;
    }

    return function (value) {
      var isUndefined = false;

      if (value === undefined) {
        isUndefined = true;
        value = {};
      } else if ((0, _componentType2.default)(value) !== 'object') {
        throw new StructError('value_invalid', { type: type, value: value });
      }

      var ret = {};

      for (var _key2 in structs) {
        var s = structs[_key2];
        var v = value[_key2];
        var r = void 0;

        try {
          r = s(v);
        } catch (e) {
          var path = [_key2].concat(e.path);

          switch (e.code) {
            case 'value_invalid':
              throw new StructError('property_invalid', _extends({}, e, { key: _key2, path: path }));
            case 'value_required':
              throw isUndefined ? new StructError('value_required', { type: type }) : new StructError('property_required', _extends({}, e, { key: _key2, path: path }));
            default:
              if ('path' in e) e.path = path;
              throw e;
          }
        }

        if (_key2 in value) {
          ret[_key2] = r;
        }
      }

      for (var _key3 in value) {
        if (!(_key3 in structs)) {
          throw new StructError('property_unknown', { key: _key3, path: [_key3] });
        }
      }

      return isUndefined ? undefined : ret;
    };
  }

  /**
   * Define a struct with `schema`.
   *
   * @param {Function|String|Array|Object} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function struct(schema, defaults) {
    var s = void 0;

    if ((0, _componentType2.default)(schema) === 'function') {
      s = schema;
    } else if ((0, _componentType2.default)(schema) === 'string') {
      s = scalarStruct(schema, defaults);
    } else if ((0, _componentType2.default)(schema) === 'array') {
      s = listStruct(schema, defaults);
    } else if ((0, _componentType2.default)(schema) === 'object') {
      s = objectStruct(schema, defaults);
    } else {
      throw new Error('A struct schema definition must be a string, array or object, but you passed: ' + schema);
    }

    return function (value) {
      if (value === undefined) {
        value = typeof defaults === 'function' ? defaults() : (0, _cloneDeep2.default)(defaults);
      }

      return s(value);
    };
  }

  /**
   * Return the struct factory.
   */

  return struct;
}

/**
 * Export the factory and the factory creator.
 *
 * @type {Function}
 */

var struct = superstruct();

exports.default = struct;
exports.struct = struct;
exports.superstruct = superstruct;
exports.StructError = StructError;